#!/bin/sh -- need to mention perl5 here to avoid recursion
'true' || eval 'exec perl5 -S $0 $argv:q';
eval '(exit $?0)' && eval 'exec perl5 -S $0 ${1+"$@"}'
& eval 'exec /usr/local/bin/perl5 -S $0 $argv:q'
    if 0;

# Get the rigth @INC path
$path=substr($0,0,rindex($0,'/'));
unshift(@INC,$path);

# Routines from my toolbox
###########################################################################
require "h3.pl";
require "h1.pl";

# Global variables
###########################################################################
$usage                  = "Usage: $0 AccessLog DatabaseDir\n";

# Databases
$sqstat_db         = "sqstat.h3";
$elapsed_db        = "elapsed.h3";
$logtags_db        = "logtags.h3";
$icp_hosts_db      = "icp-hosts.h3";
$http_hosts_db     = "http-hosts.h3";
$hierarchy_db      = "hierarchy.h3";
$family_db         = "family.h3";
$family_mb_db      = "family-mb.h3";

$http_hostcount_db = "http-hostcount.h1";
$icp_hostcount_db  = "icp-hostcount.h1";

# Misc
%l = ();
$MB = 1/(1024*1024); # It's faster to do a '* $var' than to a '/ $var'.
$log2 = 1/log(2);    # The same holds for this.
$oldts = "";
$backup = 1;

# Grouping of logtags
@tcpHITS     = ('TCP_HIT');
@tcpHITS_IMS = ('TCP_IMS_HIT', 
		'TCP_REFRESH_HIT', 
		'TCP_REF_FAIL_HIT', 
		'TCP_EXPIRED_HIT', 
		'TCP_EXP_FAIL_HIT');
@tcpREFRESH  = ('TCP_REFRESH_MISS', 
		'TCP_CLIENT_REFRESH');
@tcpMISS     = ('TCP_MISS', 
		'TCP_EXPIRED_MISS', 
		'TCP_IMS_MISS', 
		'TCP_SWAPFAIL', 
		'TCP_REFRESH',           # Old logtag
		'TCP_REFRESH_MISS', 
		'TCP_CLIENT_REFRESH');
@tcpDENY     = ('TCP_DENIED');
@tcpERR      = ('LOG_NONE', 
		'ERR_READ_TIMEOUT',
		'ERR_LIFETIME_EXP',
		'ERR_NO_CLIENTS_BIG_OBJ',
		'ERR_READ_ERROR',
		'ERR_CLIENT_ABORT',
		'ERR_CONNECT_FAIL',
		'ERR_INVALID_REQ',
		'ERR_UNSUP_REQ',
		'ERR_INVALID_URL',
		'ERR_NO_FDS',
		'ERR_DNS_FAIL',
		'ERR_NOT_IMPLEMENTED',
		'ERR_CANNOT_FETCH',
		'ERR_NO_RELAY',
		'ERR_DISK_IO',
		'ERR_ZERO_SIZE_OBJECT',
		'ERR_FTP_DISABLED',
		'ERR_PROXY_DENIED');

@udpHITS  = ('UDP_HIT', 
	     'UDP_HIT_OBJ');
@udpMISS  = ('UDP_MISS');
@udpDENY  = ('UDP_DENIED', 
	     'UDP_RELOADING');
@udpERR   = ('UDP_INVALID');

# What sould not be registered in the family.
@familySTOPLIST   = ('NONE',
		     'DIRECT',
		     'TIMEOUT_DIRECT',
		     'NO_PARENT_DIRECT',
		     'LOCAL_IP_DIRECT',
		     'FIREWALL_IP_DIRECT',
		     'NO_DIRECT_FAIL',
		     'SOURCE_FASTEST');


###########################################################################
### Main
###########################################################################

# Initiate some values
umask (027);

# Print help-information
if ($ARGV[0] eq "" || $ARGV[1] eq "") {
  printf "$usage";
  exit(0);
}
$accesslogfile = $ARGV[0];
$dbpath = $ARGV[1];

### Read the databases
###########################################################################
%sqstat           = &read_h3("$dbpath/$sqstat_db");
%elapsed          = &read_h3("$dbpath/$elapsed_db");
%logtags          = &read_h3("$dbpath/$logtags_db");
%icphosts         = &read_h3("$dbpath/$icp_hosts_db");
%httphosts        = &read_h3("$dbpath/$http_hosts_db");
%hierarchy        = &read_h3("$dbpath/$hierarchy_db");
%family           = &read_h3("$dbpath/$family_db");
%familymb         = &read_h3("$dbpath/$family_mb_db");

%httphostcount    = &read_h1("$dbpath/$http_hostcount_db");
%icphostcount     = &read_h1("$dbpath/$icp_hostcount_db");

# Fetch the logfile and count things.
###########################################################################
open (IN, $accesslogfile) || die "Could not open \"$accesslogfile\": $!\n";
while ($l = <IN>){

    # splitaccessline gives %l the following keys:
    # 
    # timestamp 
    # ts
    # elapsed
    # clientaddr
    # logtag
    # size
    # method
    # hierarchy
    # hhost

    %l = &splitaccessline($l);
    $l{size} = $l{size} * $MB;

    # Save the databases for each new day. This is done to keep track of 
    # where we are when analyzing large logfiles. This also makes data 
    # available faster. The first save happens after the first logline, 
    # and makes a backup of earlier runs.
    if ($l{ts} ne $oldts) {
	($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)=localtime(time);
	if ($sec<10) {$sec="0$sec";}
	if ($min<10) {$min="0$min";}
	if ($hour<10) {$hour="0$hour";}
	if ($mon<9) { $mon++; $mon="0$mon"; }
	if ($mday<10) {$mday="0$mday";}
	print stderr "[19$year$mon$mday $hour:$min:$sec] ";
	print stderr "Starting on $l{ts}\n";
	$oldts = $l{ts};
	&save_h3($backup, "$dbpath/$sqstat_db",       %sqstat);
	&save_h3($backup, "$dbpath/$elapsed_db",      %elapsed);
	&save_h3($backup, "$dbpath/$logtags_db",      %logtags);
	&save_h3($backup, "$dbpath/$icp_hosts_db",    %icphosts);
	&save_h3($backup, "$dbpath/$http_hosts_db",   %httphosts);
	&save_h3($backup, "$dbpath/$hierarchy_db",    %hierarchy);
	&save_h3($backup, "$dbpath/$family_db",       %family);
	&save_h3($backup, "$dbpath/$family_mb_db",    %familymb);

	&save_h1($backup, "$dbpath/$http_hostcount_db",   %httphostcount);
	&save_h1($backup, "$dbpath/$icp_hostcount_db",    %icphostcount);
	$backup = 0;
    }

    # Count things
    #########################################################################

    $logtags{$l{ts}}{A}{$l{logtag}} ++;
    $logtags{$l{ts}}{B}{$l{logtag}} += $l{size};

    if (!(grep(/^$l{hierarchy}$/, @familySTOPLIST))) {
	$family  {$l{ts}}{$l{hhost}}{$l{hierarchy}} ++;
	$familymb{$l{ts}}{$l{hhost}}{$l{hierarchy}} += $l{size};
    }

    if ($l{method} =~ /^ICP_/) { # ICP requests

	$icphostcount{$l{clientaddr}} = $l{timestamp};

	# Count number of different clients.
	if (!($icphosts{$l{ts}}{A}{$l{clientaddr}})) {
	    $sqstat{$l{ts}}{I}{C}++;
	}
	$icphosts{$l{ts}}{A}{$l{clientaddr}} ++;
	$icphosts{$l{ts}}{B}{$l{clientaddr}} += $l{size};

	$sqstat{$l{ts}}{I} {T} ++;
	$sqstat{$l{ts}}{IB}{T} += $l{size};
	if (grep(/^$l{logtag}$/, @udpHITS)) { 
	    $sqstat{$l{ts}}{I} {H} ++;
	    $sqstat{$l{ts}}{IB}{H} += $l{size};
	} elsif (grep(/^$l{logtag}$/, @udpMISS)) {
	    $sqstat{$l{ts}}{I} {M} ++;
	    $sqstat{$l{ts}}{IB}{M} += $l{size};
	} elsif (grep(/^$l{logtag}$/, @udpERR)) {
	    $sqstat{$l{ts}}{I} {E} ++;
	    $sqstat{$l{ts}}{IB}{E} += $l{size};
	} elsif (grep(/^$l{logtag}$/, @udpDENY)) {
	    $sqstat{$l{ts}}{I} {D} ++;
	    $sqstat{$l{ts}}{IB}{D} += $l{size};
	} else {
	    print stderr "UNKNOWN : $l{logtag} $l";
	}

    } else { # HTTP requests

	$httphostcount{$l{clientaddr}} = $l{timestamp};

	# Count number of different clients.
	if (!($httphosts{$l{ts}}{A}{$l{clientaddr}})) {
	    $sqstat{$l{ts}}{H}{C}++;
	}
	$httphosts{$l{ts}}{A}{$l{clientaddr}} ++;
	$httphosts{$l{ts}}{B}{$l{clientaddr}} += $l{size};

	$hierarchy{$l{ts}}{A}{$l{hierarchy}} ++;
	$hierarchy{$l{ts}}{B}{$l{hierarchy}} += $l{size};


        # Calculate the range. Can it be done faster?
	# time = 2**x => x = log(time) / log(2). I take the int for
	# this, add one and saves it. This gives me a logarithmic
	# scale where the accesstime are slottet between 2**(x+1) and 2**x.
	# This gives me a relatively small resolution on fast accesstimes, 
	# and a low resolution on slow access-times. (And it will give me
	# nice graphs. :-)
	if ($l{elapsed}>0) {
	    $range = int(log($l{elapsed})*$log2)+1;
	} else {
	    $range = 0;
	}
	$elapsed{$l{ts}}{T}{$range}++;

	$sqstat{$l{ts}}{H} {T} ++;
	$sqstat{$l{ts}}{HB}{T} += $l{size};
	if (grep(/^$l{logtag}$/, @tcpHITS)) {
	    $sqstat      {$l{ts}} {H}  {H}             ++;
	    $sqstat      {$l{ts}} {HB} {H}             += $l{size};
	    $elapsed     {$l{ts}} {H}  {$range}        ++;
	} elsif (grep(/^$l{logtag}$/, @tcpHITS_IMS)) {
	    $sqstat      {$l{ts}} {H}  {I}             ++;
	    $sqstat      {$l{ts}} {HB} {I}             += $l{size};
	    $elapsed     {$l{ts}} {I}  {$range}        ++;
	} elsif (grep(/^$l{logtag}$/, @tcpREFRESH)) {
	    $sqstat      {$l{ts}} {H}  {R}             ++;
	    $sqstat      {$l{ts}} {HB} {R}             += $l{size};
	    $elapsed     {$l{ts}} {R}  {$range}        ++;
	} elsif (grep(/^$l{logtag}$/, @tcpMISS)) {
	    $sqstat      {$l{ts}} {H}  {M}             ++;
	    $sqstat      {$l{ts}} {HB} {M}             += $l{size};
	    $elapsed     {$l{ts}} {M}  {$range}        ++;
	} elsif (grep(/^$l{logtag}$/, @tcpERR)) {
	    $sqstat      {$l{ts}} {H}  {E}             ++;
	    $sqstat      {$l{ts}} {HB} {E}             += $l{size};
	    $elapsed     {$l{ts}} {E}  {$range}        ++;
	} elsif (grep(/^$l{logtag}$/, @tcpDENY)) {
	    $sqstat      {$l{ts}} {H}  {D}             ++;
	    $sqstat      {$l{ts}} {HB} {D}             += $l{size};
	    $elapsed     {$l{ts}} {D}  {$range}        ++;
	} else {
	    print stderr "UNKNOWN : $l{logtag} $l";
	}
    }
}
close(IN);

# The final saving of the databases.
&save_h3(0, "$dbpath/$sqstat_db",       %sqstat);
&save_h3(0, "$dbpath/$elapsed_db",      %elapsed);
&save_h3(0, "$dbpath/$logtags_db",      %logtags);
&save_h3(0, "$dbpath/$icp_hosts_db",    %icphosts);
&save_h3(0, "$dbpath/$http_hosts_db",   %httphosts);
&save_h3(0, "$dbpath/$hierarchy_db",    %hierarchy);
&save_h3(0, "$dbpath/$family_db",       %family);
&save_h3(0, "$dbpath/$family_mb_db",    %familymb);

&save_h1(0, "$dbpath/$http_hostcount_db",   %httphostcount);
&save_h1(0, "$dbpath/$icp_hostcount_db",    %icphostcount);

exit(0);


########################################################################
### Subs
########################################################################


sub splitaccessline {
  local($logline) = @_;
  local(@months, @dates, $timestamp, $elapsed, $srcaddress, $typecode);
  local($size, $method, $url, $type, $code, $protocol, $tmp);
  local($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst);
  local(%tmp, %answer);

  @months = ('01','02','03','04','05','06','07','08','09','10','11','12');
  @dates  = ('01','02','03','04','05','06','07','08','09','10','11','12',
             '13','14','15','16','17','18','19','20','21','22','23','24',
             '25','26','27','28','29','30','31');

# 0) Timestamp
# 1) Elapsed Time
# 2) Client Address
# 3) Log Tag / HTTP Code
# 4) Size
# 5) Request Method
# 6) URL
# 7) Ident
# 8) Hierarchy Data / Hostname
# 9) Content Type

  chop($logline);
  @e = split(/\s+/, $logline);
  if ($#e == 9) {
      ($hierarchy, $hhost) = split(/\//, @e[8]);
  } elsif ($#e == 8) {
      ($hierarchy, $hhost) = split(/\//, @e[8]);
  } elsif ($#e > 9) { # probably an error though!
      ($hierarchy, $hhost) = split(/\//, @e[$#e-1]);
  } elsif ($#e < 8) {
      print "ERROR $#e $logline\n"; 
  }

  ($logtag, $httpcode) = split(/\//, @e[3]);
  ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)=localtime(@e[0]);
  $mon = @months[$mon]; $mday = @dates[$mday-1];
  if ($year < 10) {$year = "0$year";}
  if ($hour < 10) {$hour = "0$hour";}
  if ($min  < 10) {$min  = "0$min";}
  if ($sec  < 10) {$sec  = "0$sec";}

  $tmp{timestamp}       = @e[0];
  $tmp{ts}              = "$cent$year$mon$mday";

  $tmp{elapsed}         = @e[1];
  $tmp{clientaddr} 	= @e[2];
  $tmp{logtag} 		= $logtag;
  $tmp{size} 		= @e[4];
  $tmp{method} 		= @e[5];
  $tmp{hierarchy}       = $hierarchy;
  $tmp{hhost}           = $hhost;

  %answer = %tmp;
}
